"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.runAsWorker = exports.createSyncFn = exports.useWorkerThreads = exports.tmpdir = void 0;
const tslib_1 = require("tslib");
const child_process_1 = require("child_process");
const os_1 = require("os");
const path_1 = tslib_1.__importDefault(require("path"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const worker_threads_1 = require("worker_threads");
const uuid_1 = require("uuid");
tslib_1.__exportStar(require("./types"), exports);
/**
 * @link https://github.com/sindresorhus/temp-dir/blob/main/index.js#L9
 */
exports.tmpdir = fs_1.default.realpathSync(os_1.tmpdir());
exports.useWorkerThreads = !['0', 'false'].includes(process.env.SYNCKIT_WORKER_THREADS);
const syncFnCache = new Map();
function createSyncFn(workerPath, bufferSize) {
    if (!path_1.default.isAbsolute(workerPath)) {
        throw new Error('`workerPath` must be absolute');
    }
    const cachedSyncFn = syncFnCache.get(workerPath);
    if (cachedSyncFn) {
        return cachedSyncFn;
    }
    let resolvedWorkerPath = workerPath;
    if (!resolvedWorkerPath.endsWith('.ts')) {
        resolvedWorkerPath = require.resolve(workerPath);
    }
    const syncFn = (exports.useWorkerThreads ? startWorkerThread : startChildProcess)(resolvedWorkerPath, bufferSize);
    syncFnCache.set(workerPath, syncFn);
    return syncFn;
}
exports.createSyncFn = createSyncFn;
function startChildProcess(workerPath) {
    const executor = workerPath.endsWith('.ts') ? 'ts-node' : 'node';
    return (...args) => {
        const filename = path_1.default.resolve(exports.tmpdir, `synckit-${uuid_1.v4()}.json`);
        fs_1.default.writeFileSync(filename, JSON.stringify(args));
        const command = `${executor} ${workerPath} ${filename}`;
        try {
            child_process_1.execSync(command, {
                stdio: 'inherit',
            });
            const { result, error } = JSON.parse(fs_1.default.readFileSync(filename, 'utf8'));
            if (error) {
                throw typeof error === 'object' && error && 'message' in error
                    ? // eslint-disable-next-line unicorn/error-message
                        Object.assign(new Error(), error)
                    : error;
            }
            return result;
        }
        finally {
            fs_1.default.unlinkSync(filename);
        }
    };
}
function startWorkerThread(workerPath, bufferSize = 1024) {
    const { port1: mainPort, port2: workerPort } = new worker_threads_1.MessageChannel();
    const isTs = workerPath.endsWith('.ts');
    const worker = new worker_threads_1.Worker(isTs
        ? `require('ts-node/register');require(require('worker_threads').workerData.workerPath)`
        : workerPath, {
        eval: isTs,
        workerData: { workerPath, workerPort },
        transferList: [workerPort],
        execArgv: [],
    });
    let nextID = 0;
    const syncFn = (...args) => {
        const id = nextID++;
        const sharedBuffer = new SharedArrayBuffer(bufferSize);
        const sharedBufferView = new Int32Array(sharedBuffer);
        const msg = { sharedBuffer, id, args };
        worker.postMessage(msg);
        const status = Atomics.wait(sharedBufferView, 0, 0);
        /* istanbul ignore if */
        if (status !== 'ok' && status !== 'not-equal') {
            throw new Error('Internal error: Atomics.wait() failed: ' + status);
        }
        const { id: id2, result, error, properties, } = worker_threads_1.receiveMessageOnPort(mainPort).message;
        /* istanbul ignore if */
        if (id !== id2) {
            throw new Error(`Internal error: Expected id ${id} but got id ${id2}`);
        }
        if (error) {
            // MessagePort doesn't copy the properties of Error objects. We still want
            // error objects to have extra properties such as "warnings" so implement the
            // property copying manually.
            throw Object.assign(error, properties);
        }
        return result;
    };
    worker.unref();
    return syncFn;
}
const runAsWorker = (fn) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    if (!worker_threads_1.workerData) {
        const filename = process.argv[2];
        const content = fs_1.default.readFileSync(filename, 'utf8');
        const args = JSON.parse(content);
        let msg;
        try {
            msg = { result: (yield fn(...args)) };
        }
        catch (err) {
            msg = {
                error: err instanceof Error
                    ? { name: err.name, message: err.message, stack: err.stack }
                    : err,
            };
        }
        fs_1.default.writeFileSync(filename, JSON.stringify(msg));
        return;
    }
    /* istanbul ignore next */
    const { workerPort } = worker_threads_1.workerData;
    /* istanbul ignore next */
    worker_threads_1.parentPort.on('message', ({ sharedBuffer, id, args }) => {
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        ;
        (() => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
            const sharedBufferView = new Int32Array(sharedBuffer);
            let msg;
            try {
                msg = { id, result: (yield fn(...args)) };
            }
            catch (err) {
                const error = err;
                msg = { id, error, properties: Object.assign({}, error) };
            }
            workerPort.postMessage(msg);
            Atomics.add(sharedBufferView, 0, 1);
            Atomics.notify(sharedBufferView, 0, Number.POSITIVE_INFINITY);
        }))();
    });
});
exports.runAsWorker = runAsWorker;
//# sourceMappingURL=index.js.map